@@ -1,5 +1,5 @@
 /*
- * $Id: pam_radius_auth.c,v 1.24 2001/08/20 17:11:21 aland Exp $
+ * $Id: pam_radius_auth.c,v 1.31 2003/02/27 18:01:07 aland Exp $
  * pam_radius_auth
  *      Authenticate a user via a RADIUS session
  *
@@ -21,8 +21,32 @@
  *          requests; log more errors; add NAS-Port and NAS-Port-Type
  *          attributes to ALL packets.  Some patches based on input from 
  *          Grzegorz Paszka <Grzegorz.Paszka@pik-net.pl>
+ * 1.3.13 - Always update the configuration file, even if we're given
+ *          no options.  Patch from Jon Nelson <jnelson@securepipe.com>
+ * 1.3.14 - Don't use PATH_MAX, so it builds on GNU Hurd.
+ * 1.3.15 - Implement retry option, miscellanous bug fixes.
  *
- * See end for Copyright information */
+ *
+ *   This program is free software; you can redistribute it and/or modify
+ *   it under the terms of the GNU General Public License as published by
+ *   the Free Software Foundation; either version 2 of the License, or
+ *   (at your option) any later version.
+ *
+ *   This program is distributed in the hope that it will be useful,
+ *   but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *   GNU General Public License for more details.
+ *
+ *   You should have received a copy of the GNU General Public License
+ *   along with this program; if not, write to the Free Software
+ *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *
+ * The original pam_radius.c code is copyright (c) Cristian Gafton, 1996,
+ *                                             <gafton@redhat.com>
+ *
+ * Some challenge-response code is copyright (c) CRYPTOCard Inc, 1998.
+ *                                              All rights reserved.
+ */
 
 #define PAM_SM_AUTH
 #define PAM_SM_PASSWORD
@@ -38,6 +62,8 @@
 
 #include "pam_radius_auth.h"
 
+#define DPRINT if (ctrl & PAM_DEBUG_ARG) _pam_log
+
 /* internal data */
 static CONST char *pam_module_name = "pam_radius_auth";
 static char conf_file[BUFFER_SIZE]; /* configuration file */
@@ -94,12 +120,10 @@ static int _pam_parse(int argc, CONST char **argv, radius_conf_t *conf)
     } else if (!strcmp(*argv, "skip_passwd")) {
       ctrl |= PAM_SKIP_PASSWD;
 
-#ifdef RETRY
     } else if (!strncmp(*argv,"retry=",6)) {
       int i = atoi(*argv+6);
       i &= 0x03;		/* keep the low 3 bits only */
       ctrl |= (i << 4);
-#endif RETRY
 
     } else if (!strncmp(*argv, "client_id=", 10)) {
       if (conf->client_id) {
@@ -229,12 +253,14 @@ static int
 host2server(radius_server_t *server)
 {
   char *p;
+  int ctrl = 1; /* for DPRINT */
   
   if ((p = strchr(server->hostname, ':')) != NULL) {
     *(p++) = '\0';		/* split the port off from the host name */
   }
   
   if ((server->ip.s_addr = get_ipaddr(server->hostname)) == ((UINT4)0)) {
+    DPRINT(LOG_DEBUG, "DEBUG: get_ipaddr(%s) returned 0.\n", server->hostname);
     return PAM_AUTHINFO_UNAVAIL;
   }
 
@@ -255,16 +281,21 @@ host2server(radius_server_t *server)
       
       if (p) {			/* maybe it's not "radius" */
 	svp = getservbyname (p, "udp");
+	/* quotes allow distinction from above, lest p be radius or radacct */
+	DPRINT(LOG_DEBUG, "DEBUG: getservbyname('%s', udp) returned %d.\n", p, svp);
 	*(--p) = ':';		/* be sure to put the delimiter back */
       } else {
 	if (!server->accounting) {
 	  svp = getservbyname ("radius", "udp");
+	  DPRINT(LOG_DEBUG, "DEBUG: getservbyname(radius, udp) returned %d.\n", svp);
 	} else {
 	  svp = getservbyname ("radacct", "udp");
+	  DPRINT(LOG_DEBUG, "DEBUG: getservbyname(radacct, udp) returned %d.\n", svp);
 	}
       }
       
       if (svp == (struct servent *) 0) {
+	/* debugging above... */
 	return PAM_AUTHINFO_UNAVAIL;
       }
       
@@ -356,7 +387,7 @@ get_accounting_vector(AUTH_HDR *request, radius_server_t *server)
   MD5Init(&my_md5);
   memcpy(((char *)request) + len, server->secret, secretlen);
 
-  MD5Update(&my_md5, (char *)request, len + secretlen);
+  MD5Update(&my_md5, (unsigned char *)request, len + secretlen);
   MD5Final(request->vector, &my_md5);      /* set the final vector */
 }
 
@@ -391,7 +422,7 @@ verify_packet(char *secret, AUTH_HDR *response, AUTH_HDR *request)
    * to the secret!
    */
   if (*secret) {
-    MD5Update(&my_md5, secret, strlen(secret));
+    MD5Update(&my_md5, (unsigned char *) secret, strlen(secret));
   }
 
   MD5Final(calculated, &my_md5);      /* set the final vector */
@@ -467,8 +498,8 @@ add_password(AUTH_HDR *request, unsigned char type, CONST char *password, char *
   unsigned char misc[AUTH_VECTOR_LEN];
   int i;
   int length = strlen(password);
-  char hashed[256 + AUTH_PASS_LEN]; /* can't be longer than this */
-  char *vector;
+  unsigned char hashed[256 + AUTH_PASS_LEN]; /* can't be longer than this */
+  unsigned char *vector;
   attribute_t *attr;
 
   if (length > MAXPASS) {	/* shorten the password for now */
@@ -482,8 +513,8 @@ add_password(AUTH_HDR *request, unsigned char type, CONST char *password, char *
     length &= ~(AUTH_PASS_LEN - 1); /* chop it off */
   }                             /* 16*N maps to itself */
 
-  memset(hashed, '\0', length);
-  memcpy(hashed, password, length);
+  memset(hashed, 0, length);
+  memcpy(hashed, password, strlen(password));
 
   attr = find_attribute(request, PW_PASSWORD);
 
@@ -497,7 +528,7 @@ add_password(AUTH_HDR *request, unsigned char type, CONST char *password, char *
   /* encrypt the password */
   /* password : e[0] = p[0] ^ MD5(secret + vector) */
   MD5Init(&md5_secret);
-  MD5Update(&md5_secret, secret, strlen(secret));
+  MD5Update(&md5_secret, (unsigned char *) secret, strlen(secret));
   my_md5 = md5_secret;		/* so we won't re-do the hash later */
   MD5Update(&my_md5, vector, AUTH_VECTOR_LEN);
   MD5Final(misc, &my_md5);      /* set the final vector */
@@ -673,7 +704,7 @@ build_radius_packet(AUTH_HDR *request, CONST char *user, CONST char *password, r
     get_random_vector(request->vector);
   }
   
-  add_attribute(request, PW_USER_NAME, (char *) user, strlen(user));
+  add_attribute(request, PW_USER_NAME, (unsigned char *) user, strlen(user));
 
   /*
    *  Add a password, if given.
@@ -708,7 +739,7 @@ build_radius_packet(AUTH_HDR *request, CONST char *user, CONST char *password, r
 
   /* There's always a NAS identifier */
   if (conf->client_id && *conf->client_id) {
-    add_attribute(request, PW_NAS_IDENTIFIER, conf->client_id,
+    add_attribute(request, PW_NAS_IDENTIFIER, (unsigned char *) conf->client_id,
 		  strlen(conf->client_id));
   }
 
@@ -726,7 +757,8 @@ build_radius_packet(AUTH_HDR *request, CONST char *user, CONST char *password, r
  * Send a packet and get the response
  */
 static int
-talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response, char *password, char *old_password)
+talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response,
+	    char *password, char *old_password, int tries)
 {
   int salen, total_length;
   fd_set set;
@@ -737,10 +769,12 @@ talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response, char *pa
   struct sockaddr_in *s_in = (struct sockaddr_in *) &saremote;
   radius_server_t *server = conf->server;
   int ok;
+  int server_tries;
+  int retval;
 
   /* ************************************************************ */
   /* Now that we're done building the request, we can send it */
-  
+
   /*
     Hmm... on password change requests, all of the found server information
     could be saved with a pam_set_data(), which means even the radius_conf_t
@@ -753,11 +787,12 @@ talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response, char *pa
   
   /* loop over all available servers */
   while (server != NULL) {
-    
+
     /* only look up IP information as necessary */
-    if (host2server(server) != PAM_SUCCESS) {
-      _pam_log(LOG_ERR, "Failed looking up IP address for RADIUS server %s",
-	       server->hostname);
+    if ((retval = host2server(server)) != PAM_SUCCESS) {
+      _pam_log(LOG_ERR,
+	       "Failed looking up IP address for RADIUS server %s (errcode=%d)",
+	       server->hostname, retval);
       ok = FALSE;
       goto next;		/* skip to the next server */
     }
@@ -773,6 +808,8 @@ talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response, char *pa
       get_accounting_vector(request, server);
     }
 
+    server_tries = tries;
+send:
     /* send the packet */
     if (sendto(conf->sockfd, (char *) request, total_length, 0,
 	       &saremote, sizeof(struct sockaddr_in)) < 0) {
@@ -781,7 +818,7 @@ talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response, char *pa
       ok = FALSE;
       goto next;		/* skip to the next server */
     }
-    
+
     /* ************************************************************ */
     /* Wait for the response, and verify it. */
     salen = sizeof(struct sockaddr);
@@ -803,6 +840,8 @@ talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response, char *pa
       if (rcode == 0) {
 	_pam_log(LOG_ERR, "RADIUS server %s failed to respond",
 		 server->hostname);
+	if (--server_tries)
+	  goto send;
 	ok = FALSE;
 	break;			/* exit from the select loop */
       } else if (rcode < 0) {
@@ -814,6 +853,8 @@ talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response, char *pa
 	  if (now > end) {
 	    _pam_log(LOG_ERR, "RADIUS server %s failed to respond",
 		     server->hostname);
+	    if (--server_tries)
+	      goto send;
 	    ok = FALSE;
 	    break;			/* exit from the select loop */
 	  }
@@ -901,7 +942,7 @@ talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response, char *pa
       }
       /* otherwise, we've got data on another descriptor, keep select'ing */
     }
-    
+
     /* go to the next server if this one didn't respond */
   next:
     if (!ok) {
@@ -941,7 +982,7 @@ talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response, char *pa
       break;
     }
   }
-  
+
   if (!server) {
     _pam_log(LOG_ERR, "All RADIUS servers failed to respond.");
     return PAM_AUTHINFO_UNAVAIL;
@@ -1006,8 +1047,6 @@ static int rad_converse(pam_handle_t *pamh, int msg_style, char *message, char *
 	              , (void *) pret, _int_free );	\
 	return retval; }
 
-#define DPRINT if (ctrl & PAM_DEBUG_ARG) _pam_log
-
 PAM_EXTERN int 
 pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc,CONST char **argv)
 {
@@ -1023,17 +1062,15 @@ pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc,CONST char **argv)
   AUTH_HDR *request = (AUTH_HDR *) send_buffer;
   AUTH_HDR *response = (AUTH_HDR *) recv_buffer;
   radius_conf_t config;
-
-#ifdef RETRY
   int tries;
-#endif
 
   ctrl = _pam_parse(argc, argv, &config);
+  tries = ((ctrl & PAM_RETRY) >> 4) + 1;
 
   /* grab the user name */
   retval = pam_get_user(pamh, &user, NULL);
   PAM_FAIL_CHECK;
-  
+
   /* check that they've entered something, and not too long, either */
   if ((user == NULL) ||
       (strlen(user) > MAXPWNAM)) {
@@ -1081,12 +1118,6 @@ pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc,CONST char **argv)
     DPRINT(LOG_DEBUG, "Got password %s", password);
   }
 
-#ifdef RETRY
-  tries = (ctrl & PAM_RETRY) >> 4;
-
-retry:  
-#endif
-  
   /* no previous password: maybe get one from the user */
   if (!password) {
     if (ctrl & PAM_USE_FIRST_PASS) {
@@ -1099,8 +1130,6 @@ retry:
       retval = rad_converse(pamh, PAM_PROMPT_ECHO_OFF, "Password: ", &password);
       PAM_FAIL_CHECK;
 
-    } else {			/* skip asking for the password */
-      password = "";		/* send a NULL password */
     }
   } /* end of password == NULL */
 
@@ -1117,12 +1146,13 @@ retry:
   retval = pam_get_item(pamh, PAM_RHOST, (CONST void **) &rhost);
   PAM_FAIL_CHECK;
   if (rhost) {
-    add_attribute(request, PW_CALLING_STATION_ID, rhost, strlen(rhost));
+    add_attribute(request, PW_CALLING_STATION_ID, (unsigned char *) rhost,
+		  strlen(rhost));
   }
-  
+
   DPRINT(LOG_DEBUG, "Sending RADIUS request code %d", request->code);
 
-  retval = talk_radius(&config, request, response, password, NULL);
+  retval = talk_radius(&config, request, response, password, NULL, tries);
   PAM_FAIL_CHECK;
 
   DPRINT(LOG_DEBUG, "Got RADIUS response code %d", response->code);
@@ -1160,7 +1190,7 @@ retry:
     /* copy the state over from the servers response */
     add_attribute(request, PW_STATE, a_state->data, a_state->length - 2);
 
-    retval = talk_radius(&config, request, response, resp2challenge, NULL);
+    retval = talk_radius(&config, request, response, resp2challenge, NULL, 1);
     PAM_FAIL_CHECK;
 
     DPRINT(LOG_DEBUG, "Got response to challenge code %d", response->code);
@@ -1170,15 +1200,6 @@ retry:
   if (response->code == PW_AUTHENTICATION_ACK) {
     retval = PAM_SUCCESS;
   } else {
-#ifdef RETRY
-    if (--tries > 0) {
-      /* don't do this for challenge-response users! */
-      _pam_forget(password);
-      request->id += 16;		/* don't use the same RADIUS id */
-      goto retry;		/* nasty, nasty. */
-    }
-#endif
-
     retval = PAM_AUTH_ERR;	/* authentication failure */
 
   error:
@@ -1282,7 +1303,8 @@ pam_private_session(pam_handle_t *pamh, int flags,
   add_int_attribute(request, PW_ACCT_STATUS_TYPE, status);
 
   sprintf(recv_buffer, "%08d", (int) getpid());
-  add_attribute(request, PW_ACCT_SESSION_ID, recv_buffer, strlen(recv_buffer));
+  add_attribute(request, PW_ACCT_SESSION_ID, (unsigned char *) recv_buffer,
+		strlen(recv_buffer));
 
   add_int_attribute(request, PW_ACCT_AUTHENTIC, PW_AUTH_RADIUS);
 
@@ -1292,7 +1314,7 @@ pam_private_session(pam_handle_t *pamh, int flags,
     add_int_attribute(request, PW_ACCT_SESSION_TIME, time(NULL) - session_time);
   }
 
-  retval = talk_radius(&config, request, response, NULL, NULL);
+  retval = talk_radius(&config, request, response, NULL, NULL, 1);
   PAM_FAIL_CHECK;
 
   /* oops! They don't have the right password.  Complain and die. */
@@ -1411,7 +1433,7 @@ pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, CONST char **argv)
     build_radius_packet(request, user, password, &config);
     add_int_attribute(request, PW_USER_SERVICE_TYPE, PW_AUTHENTICATE_ONLY);
     
-    retval = talk_radius(&config, request, response, password, NULL);
+    retval = talk_radius(&config, request, response, password, NULL, 1);
     PAM_FAIL_CHECK;
 
     /* oops! They don't have the right password.  Complain and die. */
@@ -1514,7 +1536,7 @@ pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, CONST char **argv)
     build_radius_packet(request, user, new_password, &config);
     add_password(request, PW_OLD_PASSWORD, password, password);
     
-    retval = talk_radius(&config, request, response, new_password, password);
+    retval = talk_radius(&config, request, response, new_password, password, 1);
     PAM_FAIL_CHECK;
 
     /* Whew! Done password changing, check for password acknowledge */
@@ -1554,6 +1576,18 @@ pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, CONST char **argv)
   return retval;
 }
 
+/*
+ *  Do nothing for account management.  This is apparently needed by
+ *  some programs.
+ */
+PAM_EXTERN int
+pam_sm_acct_mgmt(pam_handle_t *pamh,int flags,int argc,CONST char **argv)
+{
+  int retval;
+  retval = PAM_SUCCESS;
+  return retval;
+}
+
 #ifdef PAM_STATIC
 
 /* static module data */
@@ -1562,54 +1596,10 @@ struct pam_module _pam_radius_modstruct = {
   "pam_radius_auth",
   pam_sm_authenticate,
   pam_sm_setcred,
-  NULL,
+  pam_sm_acct_mgmt,
   pam_sm_open_session,
   pam_sm_close_session,
   pam_sm_chauthtok,
 };
 #endif
 
-/*
- *  This module is a merger of an old version of pam_radius.c, and
- * code which went into mod_auth_radius.c, with further modifications
- * by Alan DeKok of CRYPTOCard Inc..
- *
- * The original pam_radius.c code is copyright (c) Cristian Gafton, 1996,
- *                                             <gafton@redhat.com>
- *
- * The additional code is copyright (c) CRYPTOCard Inc, 1998.
- *
- *                                              All rights reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, and the entire permission notice in its entirety,
- *    including the disclaimer of warranties.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote
- *    products derived from this software without specific prior
- *    written permission.
- * 
- * ALTERNATIVELY, this product may be distributed under the terms of
- * the GNU Public License, in which case the provisions of the GPL are
- * required INSTEAD OF the above restrictions.  (This clause is
- * necessary due to a potential bad interaction between the GPL and
- * the restrictions contained in a BSD-style copyright.)
- * 
- * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
- * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
- * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
- * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
- * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
- * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
- * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
- * OF THE POSSIBILITY OF SUCH DAMAGE.
- *
- */
